#include <sys/fcntl.h>
#include <sys/stat.h>

#ifdef WIN32
# include <winsock2.h>
# include <process.h>
# include <windows.h>
#else
# include <sys/poll.h>
# include <sys/socket.h>
# include <netdb.h>
# include <netinet/in.h>
# include <arpa/inet.h> /* for inet_ntoa */
# include <sys/wait.h>
# include <sys/socket.h>
#endif

#include <signal.h>
#include <unistd.h>

#include <vector>
#include <cerrno>
#include <cstdio>
#include <set>
#include <utility>

#include "config.hh"
#include "cset.hh"

/* Configuration */

static const Chain2& FindUTF8()
{
    static const Chain2 utf8 = CreateBuiltinChain("utf-8");
    return utf8;
}
static const Chain2& FindChain2For(const std::string& channame)
{
    static class ChannelCache
    {
        std::map<std::string, Chain2> cache;
    public:
        ChannelCache()
        {
            ConfParser::FieldType specials = GetConf("settings", "special_chain");
            for(unsigned a=0; a<specials.size(); ++a)
            {
                const ConfParser::FuncType& func = specials[a];
                
                if(func.size() <= 1)
                {
                    continue;
                }
                
                try
                {
                    Chain2 chain2 = FindChain2(func[0].SField);
                    for(unsigned b=1; b<func.size(); ++b)
                    {
                        cache[func[b].SField] = chain2;
                    }
                }
                catch(const std::string& chainname)
                {
                    fprintf(stderr, "Chain '%s' is not defined, although referenced in special_chain settings.\n",
                        chainname.c_str());
                }
            }
        }
        const Chain2& FindForChannel(const std::string& channame) const
        {
            std::map<std::string, Chain2>::const_iterator i;
            i = cache.find(channame);

            if(i == cache.end())
            {
                static const std::string defaultname = GetConf("settings", "default_chain").SField();

                try
                {
                    static const Chain2& Default = FindChain2(defaultname);
                    return Default;
                }
                catch(const std::string& chainname)
                {
                    fprintf(stderr, "Default chain '%s' not defined! Falling back to UTF8 default\n",
                        defaultname.c_str());
                    return FindUTF8();
                }
            }
            return i->second;
        }
    } ChannelCache;
    
    return ChannelCache.FindForChannel(channame);
}
static const std::string GetServer()
{
    return GetConf("settings", "server_host").SField();
}
static const unsigned GetServerPort()
{
    return GetConf("settings", "server_port").IField();
}
static const unsigned GetListenPort()
{
    return GetConf("settings", "listen_port").IField();
}

/* Util */

const std::vector<std::string> Split(const std::string& text, char separator)
{
    std::vector<std::string> words;
    std::string rest = text;
    for(;;)
    {
        unsigned spacepos = rest.find(separator);  
        if(spacepos == rest.npos)break;
        words.push_back(rest.substr(0, spacepos));   
        if(spacepos+1 >= rest.size()) { rest = ""; break; }
        rest = rest.substr(spacepos+1);
    }
    if(rest.size()) { words.push_back(rest); rest = ""; }
    return words;
}


/* From client(utf8) to server */
static const std::string OutChannel(const std::string& chan, const std::string& line)
{
    const Chain2& chain = FindChain2For(chan);
    const Chain2& utf8  = FindUTF8();
    
    return chain.out.Encode(utf8.in.Decode(line));
}


/* From server to client(utf8) */
static const std::string InChannel(const std::string& chan, const std::string& line)
{
    const Chain2& chain = FindChain2For(chan);
    const Chain2& utf8  = FindUTF8();
    
    return utf8.out.Encode(chain.in.Decode(line));
}

class SockStream
{
#ifndef WIN32
    mutable FILE* fp;
#else
    mutable int fd;
#endif
public:
    SockStream()
    {
#ifndef WIN32
        fp=NULL;
#else
        fd=-1;
#endif
    }
    SockStream(int sock)
    {
#ifndef WIN32
        // Nonblocking is not wanted
        fcntl(sock, F_SETFL,  fcntl(sock, F_GETFL) & ~O_NONBLOCK);
        fp = fdopen(sock, "r+");
        // Buffering is not wanted
        setbuf(fp,  NULL);
#else
        fd=sock;
#endif
    }
    SockStream(const SockStream& b)
    {
#ifndef WIN32
        fp=b.fp; b.fp=NULL;
#else
        fd=b.fd; b.fd=-1; 
#endif
    }
    ~SockStream()
    {
#ifndef WIN32
        if(fp)fclose(fp);
#else
        if(fd>=0)closesocket(fd);
#endif
    }
    SockStream& operator =(const SockStream& b)
    {
#ifndef WIN32
        fp=b.fp; b.fp=NULL;
#else
        fd=b.fd; b.fd=-1; 
#endif
        return *this;
    }
    const std::string ReadLine()
    {
#ifndef WIN32
        char Buf[8192];
        if(!fgets(Buf, sizeof Buf, fp)) return "";
        return Buf;
#else
        std::string result;
        for(;;)
        {
            char byte;
            int i = recv(fd, &byte, 1, 0);
            if(i == 1) { result += byte; if(byte == '\n') break; continue; }
            break;
        }
        return result;
#endif
    }
    void Write(const std::string& Buf)
    {
#ifndef WIN32
        fwrite(Buf.c_str(), 1, Buf.size(), fp);
#else
        send(fd, Buf.c_str(), Buf.size(), 0);
#endif
    }
    int GetFD() const
    {
#ifndef WIN32
        return fileno(fp);
#else
        return fd;
#endif
    }
};

static bool ReadWrite(SockStream& inf, SockStream& outf)
{
    std::string line = inf.ReadLine();
    if(line.empty()) return true;
    
    std::string guess_channel = "";
    
    enum { in, out } guess_dir = out;

    std::vector<std::string> words = Split(line, ' ');
    for(unsigned a=0; a<4; ++a)
    {
        if(a >= words.size()) break;
        const std::string& word = words[a];
        if(word.empty()) continue;
        
        char ch = word[0];
        if(a == 0 && ch == ':')
        {
            guess_dir = in; // from server
        }
        if(ch == '#' || ch == '&' || ch == '+')
        {
            guess_channel = word;
            break;
        }
    }
    
/*
    fprintf(stderr, "guess(chan=%s, dir=%s)\n",
        guess_channel.c_str(),
        guess_dir==in ? "in" : "out");
*/
    std::string result =
        guess_dir == in ?  InChannel(guess_channel, line)
                        : OutChannel(guess_channel, line);
    
    outf.Write(result);
    return false;
}

static SockStream ConnectIRC()
{
    struct sockaddr_in MyAddr;
    
    struct hostent *HostName = gethostbyname(GetServer().c_str());
    if(!HostName)
    {
        throw "ConnectIRC::gethostbyname";
    }

    memset(&MyAddr, 0, sizeof(MyAddr));

    MyAddr.sin_family = AF_INET;
    MyAddr.sin_port = htons(GetServerPort());
    memcpy(&MyAddr.sin_addr, HostName->h_addr, HostName->h_length);

    int Plug = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(Plug < 0)
    {
        throw "ConnectIRC::socket";
    }
    
    if(connect(Plug, (struct sockaddr*)&MyAddr, sizeof(MyAddr)) == -1)
    {
        close(Plug);
        throw "ConnectIRC::connect";
    }
    
    return SockStream(Plug);
}

void ReportSocketError(const char* msg)
{
#ifndef WIN32
    perror(msg);
#else
    int err = errno;
    printf("%s: error %d ", msg, WSAGetLastError());
    printf("'%s'\n", sys_errlist[err]);
    //WSACleanUp();
#endif
}


#ifndef WIN32
/* POSIX version */
static void ClientLoop(SockStream& user, SockStream& server)
{
    for(;;)
    {
        struct pollfd polls[2] =
        {
            {user.GetFD(), POLLIN, 0},
            {server.GetFD(), POLLIN, 0}
        };
        int n = poll(polls, 2, -1);
        if(n < 0)
        {
            if(errno == EINTR) continue;
            break;
        }
        if(polls[0].revents & POLLIN)
        {
            bool eof = ReadWrite(user, server);
            if(eof) break;
        }
        if(polls[1].revents & POLLIN)
        {
            bool eof = ReadWrite(server, user);
            if(eof) break;
        }
    }
}
#else
/* Windows version */
static void ClientLoop(SockStream& user, SockStream& server)
{
    for(;;)
    {
        int u = user.GetFD();
        int s = server.GetFD();
        int upper = u>s ? u : s;
        
        fd_set r;
        FD_ZERO(&r);
        FD_SET(u, &r);
        FD_SET(s, &r);
        
        int n = select(upper+1, &r, NULL, NULL, NULL);
        if(n < 0)
        {
            if(errno == EINTR) continue;
            break;
        }
        if(FD_ISSET(u, &r))
        {
            bool eof = ReadWrite(user, server);
            if(eof) break;
        }
        if(FD_ISSET(s, &r))
        {
            bool eof = ReadWrite(server, user);
            if(eof) break;
        }
    }
}
#endif

static void Service(SockStream user)
{
    SockStream server;
    try
    {
        server = ConnectIRC();
    }
    catch(const char *s)
    {
        char Buf[512];
        
        std::string error = "Can not connect to ";
        error += GetServer();
        error += ':';
        sprintf(Buf, "%u", GetServerPort());
        error += Buf;
        
        error += ": ";
        error += s;
        error += ": ";
        if(errno != 0)
        {
            error += strerror(errno);
        }
        
        error = "ERROR :" + error + "\n";
        user.Write(error);
        return;
    }
    
    ClientLoop(user, server);
}

#ifndef WIN32
/* POSIX version */
static void CatchDead(int);
static void LaunchSocketHandler(int client, int listener)
{
    /* Listen to child deaths */
    signal(SIGCHLD, CatchDead);

    int forkpid = fork();
    switch(forkpid)
    {
        case -1:
        {
            /* Failed? */
            perror("fork");
            break;
        }
        case 0:
        {
            /* This starts a child process. */
            
            /* Child closes his unneeded resources. */
            close(listener);
            close(STDERR_FILENO);
            /* Child doesn't except child deaths. */
            signal(SIGCHLD, SIG_IGN);
            /* Child starts handling the client. */
            Service(client);
            /* Client has terminated. End. */
            _exit(0);
        }
    }
    /* In any case, the main process doesn't
     * need to worry about the client anymore.
     */
    close(client);
}
static void CatchDead(int)
{
    /* Wait for the dead. */
    waitpid(-1, NULL, WNOHANG);
}
#else
/* Windows version */
static DWORD WINAPI WindowsClientThread(LPVOID param)
{
    Service((int)param);
    return 0;
}
static void LaunchSocketHandler(int client, int listener)
{
    listener=listener; /* unused */
    DWORD threadid;
    CloseHandle(CreateThread
        (NULL, 0, WindowsClientThread,
         (LPVOID)(client),
         0,
         &threadid));
}
#endif


int main(int argc, char**argv)
{
#ifdef WIN32
    WSADATA wsaData;
    if(WSAStartup(MAKEWORD(2,2),&wsaData) == SOCKET_ERROR)
    {
        fprintf(stderr,"WSAStartup failed with error %d\n", WSAGetLastError());
        WSACleanup();
        return -1;
    }
#endif
    printf(
"iconvircproxy version "VERSION"\n"
"Copyright (C) 1992,2004 Joel Yliluoma (http://iki.fi/bisqwit/)\n"
"\n"
"iconvircproxy is free software, covered by the General Public License,\n"
"and you are welcome to change it and/or distribute copies of it under\n"
"certain conditions. Read the COPYING file to see the conditions.\n"
"There is absolutely no warranty for iconvircproxy.\n");

    /* Close everything but stderr */
#ifndef WIN32
    for(int fd=0;fd<512;++fd) { if(fd != STDERR_FILENO) close(fd); }
#endif
    struct sockaddr_in MyAddr, They;
    
    memset(&MyAddr, 0, sizeof(MyAddr));

    MyAddr.sin_family = AF_INET;
    MyAddr.sin_port = htons(GetListenPort());
    MyAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* auto-fill with my IP, network long  */

    int Plug = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(Plug < 0) { ReportSocketError("main::socket"); return -errno; }
    
#ifndef WIN32
    /* Win32 doesn't support setsockopt? */
    int flag=1; /* Just a temp variable, boolean value for setsockopt() */
    setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof flag);
#endif

    if(bind(Plug, (struct sockaddr *)&MyAddr, sizeof(struct sockaddr)))
    {
        ReportSocketError("main::bind");
        close(Plug);
        return -errno;
    }
    
    if(listen(Plug, 10))
    {
        close(Plug);
        ReportSocketError("main::listen");
        return -errno;
    }
    
    /* Flush error messages from undefined charsets. */
    FindChain2For("");
    
    printf("%s started successfully. Listening to connections.\n", argv[0]);
    for(;;)
    {
#ifndef WIN32
        unsigned AddrLen = sizeof(They);
#else
        int AddrLen = sizeof(They);
#endif
        int Sock = accept(Plug, (struct sockaddr*)&They, &AddrLen);
        if(Sock < 0)
        {
            if(errno == EINTR || errno == EAGAIN) continue;
            
            ReportSocketError("accept");
            break;
        }
        
        LaunchSocketHandler(Sock, Plug);
    }
    close(Plug);
    return 0;
}
